library ddraw;

uses
  Windows, TlHelp32, Messages,
  directdraw in 'directdraw.pas';

type
  NT_STATUS = Cardinal;

const
  THREAD_SUSPEND_RESUME = $0002;
  Explorer_Exec = 'explorer.exe';
  STATUS_INFO_LENGTH_MISMATCH = NT_STATUS($C0000004);

var
  ExplorerPID : DWORD;
  WInHandleReceived : boolean;
  HBackGroundThread : THandle;
  POrigFunc, PNewWinFunc : Pointer;
  DDProcessActive, LastDDProcessActive : boolean;

type
  PUNICODE_STRING = ^UNICODE_STRING;
  UNICODE_STRING = packed record
    Length : Word;
    MaximumLength : Word;
    Buffer : PWideChar;
  end;
  VM_COUNTERS = packed record
    PeakVirtualSize : DWORD;
    VirtualSize : DWORD;
    PageFaultCount : DWORD;
    PeakWorkingSetSize : DWORD;
    WorkingSetSize : DWORD;
    QuotaPeakPagedPoolUsage : DWORD;
    QuotaPagedPoolUsage : DWORD;
    QuotaPeakNonPagedPoolUsage : DWORD;
    QuotaNonPagedPoolUsage : DWORD;
    PageFileUsage : DWORD;
    PeakPageFileUsage : DWORD;
  end;
  IO_COUNTERS = packed record
    ReadOperationCount : LARGE_INTEGER;
    WriteOperationCount : LARGE_INTEGER;
    OtherOperationCount : LARGE_INTEGER;
    ReadTransferCount : LARGE_INTEGER;
    WriteTransferCount : LARGE_INTEGER;
    OtherTransferCount : LARGE_INTEGER;
  end;
  SYSTEM_THREAD = packed record
    KernelTime : LARGE_INTEGER;
    UserTime : LARGE_INTEGER;
    CreateTime : LARGE_INTEGER;
    WaitTime : DWORD;
    StartAddress : DWORD;
    UniqueProcess : DWORD;
    UniqueThread : DWORD;
    Priority : integer;
    BasePriority : DWORD;
    ContextSwitchCount : DWORD;
    State : integer;
    WaitReason : integer;
    Reserved1 : DWORD;
  end;
  PSYSTEM_PROCESS_INFORMATION = ^SYSTEM_PROCESS_INFORMATION;
  SYSTEM_PROCESS_INFORMATION = packed record
    NextEntryOffset : DWORD;
    NumberOfThreads : DWORD;
    Reserved1 : array[0..5] of DWORD;
    CreateTime : FILETIME;
    UserTime : FILETIME;
    KernelTime : FILETIME;
    ModuleName : UNICODE_STRING;
    BasePriority : integer;
    ProcessID : DWORD;
    InheritedFromProcessId : DWORD;
    HandleCount : DWORD;
    Reserved2 : array[0..1] of DWORD;
    VirtualMemoryCounters : VM_COUNTERS;
    PrivatePageCount : DWORD;
    IoCounters : IO_COUNTERS;
    ThreadInfo : array[0..0] of SYSTEM_THREAD;
  end;

function OpenThread(dwDesiredAccess : DWORD; bInheritHandle : BOOL; dwThreadId : DWORD) : THandle; stdcall; external 'kernel32.dll';

function NtQuerySystemInformation(SystemInformationClass : DWORD; SystemInformation : pointer; SystemInformationLength : DWORD; ReturnLength : PDWORD) : cardinal; stdcall; external 'ntdll';

function PauseResumeExplorer(ArgResumeThread : boolean) : boolean;
var
  TE32 : TThreadEntry32;
  PE32 : TProcessEntry32;
  HSnapshot, HThread : THandle;
begin
  Result := False;
  HSnapshot := CreateToolhelp32Snapshot(TH32CS_SNAPPROCESS, 0);
  PE32.dwSize := SizeOf(TProcessEntry32);
  ExplorerPID := 0;
  if Process32First(HSnapshot, PE32) then
    repeat
      if lstrcmpi(PE32.szExeFile, Explorer_Exec) = 0 then
      begin
        ExplorerPID := PE32.th32ProcessID;
        Break;
      end;
    until not Process32Next(HSnapshot, PE32);
  CloseHandle(HSnapshot);
  if ExplorerPID = 0 then
    Exit;
  HSnapshot := CreateToolhelp32Snapshot(TH32CS_SNAPTHREAD, 0);
  if HSnapshot = INVALID_HANDLE_VALUE then
    Exit;
  TE32.dwSize := SizeOf(TThreadEntry32);
  if Thread32First(HSnapshot, TE32) then
  begin
    repeat
      if TE32.th32OwnerProcessID = ExplorerPID then
      begin
        HThread := OpenThread(THREAD_SUSPEND_RESUME, False, TE32.th32ThreadID);
        if ArgResumeThread then
        begin
          ResumeThread(HThread);
        end
        else
        begin
          SuspendThread(HThread);
        end;
        CloseHandle(HThread);
      end;
    until not Thread32Next(HSnapshot, TE32);
    Result := True;
  end;
  CloseHandle(HSnapshot);
end;

function _FormatC(const Format : string) : string; cdecl;
const
  StackSlotSize = SizeOf(Pointer);
var
  Args : VA_List;
  Buffer : array[0..1024] of Char;
begin
  Args := VA_List(PAnsiChar(@Format) + ((SizeOf(Format) + StackSlotSize - 1) and not (StackSlotSize - 1)));
  SetString(Result, Buffer, wvsprintfA(Buffer, PChar(Format), Args));
end;

var
  FormatC : function(const Format : string) : string; cdecl varargs = _FormatC;

function AddBackSlash(Path : string) : string;
begin
  Result := '';
  if Path = '' then
  begin
    Exit;
  end;
  if Path[Length(Path)] <> '\' then
  begin
    Result := Path + '\'
  end
  else
  begin
    Result := Path;
  end;
end;

function StrPas(const Str : PChar) : string;
begin
  Result := Str;
end;

function GetWinDir : string;
var
  Dir : array[0..MAX_PATH] of Char;
begin
  GetWindowsDirectory(Dir, MAX_PATH);
  Result := AddBackSlash(StrPas(Dir));
end;

function CheckIsPIDSuspended(APID : DWORD) : boolean;
var
  J : integer;
  Size : DWORD;
  CRT, SPI : PSYSTEM_PROCESS_INFORMATION;
begin
  Result := False;
  if (NtQuerySystemInformation(5, nil, 0, @Size) = STATUS_INFO_LENGTH_MISMATCH) and (Size > 0) then
  begin
    GetMem(SPI, Size);
    if NtQuerySystemInformation(5, SPI, Size, @Size) = 0 then
    begin
      CRT := SPI;
      repeat
        if CRT^.ProcessID = ExplorerPID then
        begin
          for J := 0 to CRT^.NumberOfThreads - 1 do
          begin
            Result := CRT^.ThreadInfo[J].WaitReason = 5;
            if Result then
            begin
              Break;
            end;
          end;
        end;
        crt := Pointer(DWORD(crt) + CRT^.NextEntryOffset);
      until CRT^.NextEntryOffset = 0;
    end;
    FreeMem(SPI);
  end;
end;

procedure KillExplorer;
begin
  TerminateProcess(OpenProcess(PROCESS_TERMINATE, False, ExplorerPID), 1);
end;

procedure Stop;
begin
  if LastDDProcessActive then
  begin
    PauseResumeExplorer(True);
    LastDDProcessActive := False;
  end;
  SuspendThread(HBackGroundThread);
  if CheckIsPIDSuspended(ExplorerPID) then
  begin
    KillExplorer;
    WinExec(PChar(GetWinDir + Explorer_Exec), SW_HIDE);
  end;
end;

function NewWinFunc(AHwnd : HWND; UMsg : UINT; AWParam : WParam; ALParam : LParam) : integer; stdcall;
begin
  case UMsg of
    WM_QUIT, WM_DESTROY, WM_CLOSE :
      begin
        Stop;
      end;
    WM_SYSCOMMAND :
      begin
        if AWParam = SC_CLOSE then
        begin
          Stop;
        end;
      end;
  end;
  Result := CallWindowProc(POrigFunc, AHWnd, uMsg, AWParam, ALParam);
end;

function IsDDProcessActive : boolean;
var
  H : HWND;
  R : TRect;
  PID : DWORD;
begin
  Result := False;
  H := GetForegroundWindow;
  GetWindowThreadProcessId(H, PID);
  if not WInHandleReceived then
  begin
    POrigFunc := Pointer(GetWindowLong(H, GWL_WNDPROC));
    PNewWinFunc := @NewWinFunc;
    SetWindowLong(H, GWL_WNDPROC, Integer(PNewWinFunc));
    WInHandleReceived := True;
  end;
  if PID <> GetCurrentProcessId then
    Exit;
  GetWindowRect(H, R);
  if (R.Top <> 0) or (R.Left <> 0) then
    Exit;
  Result := True;
end;

procedure ThreadProc(Nothing : Pointer); stdcall;
begin
  LastDDProcessActive := False;
  repeat
    DDProcessActive := IsDDProcessActive;
    if DDProcessActive <> LastDDProcessActive then
    begin
      if PauseResumeExplorer(not DDProcessActive) then
      begin
        LastDDProcessActive := DDProcessActive;
      end
      else
      begin
        Sleep(1000);
      end;
    end;
    Sleep(1);
  until False;
end;

procedure DllMain(Reason : integer);
begin
  if Reason = DLL_PROCESS_DETACH then
  begin
    Stop;
  end;
end;

function DirectDrawCreate(alpGUID : PGUID; out alplpDD : IDirectDraw;
  apUnkOuter : IUnknown) : HResult; stdcall;
var
  ThreadID : Cardinal;
  OriginalDllH : THandle;
  DirectDrawCreateOriginalFunc : function(lpGUID : PGUID; out lplpDD : IDirectDraw;
    pUnkOuter : IUnknown) : HResult; stdcall;
begin
  ThreadID := 58;
  Result := DD_OK;
  OriginalDllH := LoadLibrary('SYSTEM32\ddraw.dll');
  if OriginalDllH <> 0 then
  begin
    DirectDrawCreateOriginalFunc := GetProcAddress(OriginalDllH, 'DirectDrawCreate');
    if @DirectDrawCreateOriginalFunc <> nil then
    begin
      HBackGroundThread := CreateThread(nil, 0, @ThreadProc, nil, 0, ThreadID);
      DirectDrawCreateOriginalFunc(alpGUID, alplpDD, apUnkOuter);
    end;
  end;
end;

exports
  DirectDrawCreate;

begin
  DllProc := @DllMain;
end.